/**
 * Copyright 2009 Atsuya Takagi
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */
package edu.csun.ecs.cs.multitouchj.ui.graphic.image;

import java.io.File;
import java.net.URL;
import java.util.TreeMap;

import edu.csun.ecs.cs.multitouchj.utility.FileUtility;

/**
 * @author Atsuya Takagi
 *
 * $Id: TextureManager.java 82 2009-03-15 23:07:07Z Atsuya Takagi $
 */
public class TextureManager {
    private static TextureManager instance = null;
    private ImageHandler imageHandler;
    private TextureHandler textureHandler;
    private TreeMap<String, Texture> textures;
    
    
    protected TextureManager() {
        imageHandler = new ImageHandlerImageIo();
        textureHandler = new TextureHandler();
        textures = new TreeMap<String, Texture>();
    }
    
    public static void create() {
        if(instance == null) {
            instance = new TextureManager();
        }
    }
    
    public static TextureManager getInstance() {
        return instance;
    }
    
    public Texture createTexture(URL url) throws Exception {
        return createTexture(url, false);
    }
    
    public Texture createTexture(URL url, boolean createNew) throws Exception {
        Texture texture = (createNew) ? null : textures.get(url.toString());
        if(texture == null) {
            File file = new File(url.getPath());
            String fileExtension = FileUtility.getExtension(file.getName());
            if(imageHandler.isSupported(fileExtension)) {
                Image image = imageHandler.decode(url);
                texture = createAndRegisterTexture(url.toString(), image);
            } else {
                throw new Exception("File type not supported: "+url.toString());
            }
        }
        
        return texture;
    }
    
    public Texture createTexture(String name, Image image) throws Exception {
        return createTexture(name, image, false);
    }
    
    public Texture createTexture(String name, Image image, boolean createNew) throws Exception {
        Texture texture = (createNew) ? null : textures.get(name);
        if(texture == null) {
            texture = createAndRegisterTexture(name, image);
        }
        
        return texture;
    }
    
    public void destroy() {
        synchronized(textures) {
            for(Texture texture : textures.values()) {
                textureHandler.deleteTexture(texture);
            }
            textures.clear();
        }
    }
    
    protected Texture createAndRegisterTexture(String name, Image image) {
        Texture texture = textureHandler.generateTexture(image);
        texture.getImage().emptyData();
        
        synchronized(textures) {
            if(textures.containsKey(name)) {
                textureHandler.deleteTexture(textures.get(name));
            }
            textures.put(name, texture);
        }
        
        return texture;
    }
}
